"""
Module defining cell mechanisms
"""
from __future__ import absolute_import
from ._neuron import Neuron
from ..utils import ConfigT


class Mechanism(ConfigT):
    _mec_name = None

    # Declare the subtypes
    HH = None   # type: HH
    PAS = None  # type: PAS

    def __new__(cls, **opts):
        if cls is Mechanism:
            raise TypeError("Mechanisms is abstract. Instantiate a specific Mechanism")
        return object.__new__(cls)

    def _apply_f(self, section, opts_dict):
        section.insert(self._mec_name)
        for key, val in opts_dict.items():
            setattr(section, "{}_{}".format(key, self._mec_name), val)

    def apply(self, obj_or_list, **kw):
        if not isinstance(obj_or_list, Neuron.Section):
            if hasattr(obj_or_list, "__iter__"):
                obj_or_list = tuple(iter(obj_or_list))
            else:
                raise TypeError("Object {} cant be assigned Mechanism properties")
        ConfigT.apply(self, obj_or_list)


class HH(Mechanism):
    _mec_name = "hh"
    gnabar = None  # 0.120 mho/cm2   Maximum specific sodium channel conductance
    gkbar  = None  # 0.036 mho/cm2   Maximum potassium channel conductance
    gl     = None  # 0.0003 mho/cm2  Leakage conductance
    el     = None  # -54.3 mV        Leakage reversal potential
    m      = None  # ?               sodium activation state variable
    h      = None  # ?               sodium inactivation state variable
    n      = None  # ?               potassium activation state variable
    ina    = None  # mA/cm2          sodium current through the hh channels
    ik     = None  # mA/cm2          potassium current through the hh channels


class PAS(Mechanism):
    _mec_name = "pas"
    g = None  # mho/cm2      conductance
    e = None  # mV           reversal potential
    i = None  # mA/cm2       non-specific current


Mechanism.HH = HH
Mechanism.PAS = PAS



#!/usr/bin/env python
"""
An example on how node can be used to mimick neurodamus behavior
"""
from __future__ import print_function
from neurodamus import Node, Neurodamus
from neurodamus.core import NeurodamusCore as Nd
from neurodamus.utils import setup_logging
import sys
import logging
from os import path as Path

RECIPE_FILE = Path.expanduser("~/dev/TestData/build/circuitBuilding_1000neurons/BlueConfig")
DEFAULT_LOG_LEVEL = 2
TRACE_LOG_LEVEL = 5


def test_run():
    """A Neurodamus typical run can be quickly setup and run using the Neurodamus class
    """
    Neurodamus(RECIPE_FILE).run()


def test_node_run(trace=False):
    """Node is more of a low-level class, where all initialization steps are manual
    """
    setup_logging(DEFAULT_LOG_LEVEL)
    if trace:
        # Some additional logging is available at special level 5
        print("TRACE mode is ON")
        logging.root.setLevel(TRACE_LOG_LEVEL)

    node = Node(RECIPE_FILE)
    node.load_targets()
    node.compute_load_balance()
    node.create_cells()
    node.execute_neuron_configures()

    logging.info("Create connections")
    node.create_synapses()
    node.create_gap_junctions()

    logging.info("Enable Stimulus")
    node.enable_stimulus()

    logging.info("Enable Modifications")
    node.enable_modifications()

    if trace:
        logging.info("Dumping config")
        Nd.stdinit()
        node.dump_circuit_config("")

    logging.info("Enable Reports")
    node.enable_reports()

    logging.info("Run")
    node.run(True)

    logging.info("Simulation finished. Gather spikes then clean up.")
    node.spike2file("out.dat")
    node.cleanup()


if __name__ == "__main__":
    test_node_run("trace" in sys.argv)



"""
Module implementing entry functions
"""
from __future__ import absolute_import

import logging
import os
import sys
import time
from docopt import docopt
from os.path import abspath
from pathlib import Path

from . import Neurodamus
from .core import MPI, OtherRankError
from .core.configuration import ConfigurationError, LogLevel, EXCEPTION_NODE_FILENAME
from .hocify import Hocify, process_file as hocify_process_file
from .utils.pyutils import docopt_sanitize


def neurodamus(args=None):
    """neurodamus

    Usage:
        neurodamus <ConfigFile> [options]
        neurodamus --help

    Options:
        -v --verbose            Increase verbosity level
        --debug                 Extremely verbose mode for debugging
        --disable-reports       Disable all reports [default: False]
        --build-model=[AUTO, ON, OFF]
                                Shall it build and eventually overwrite? [default: AUTO]
                                - AUTO: build model if doesn't exist and simulator is coreneuron
                                - ON: always build and eventually overwrite the model
                                - OFF: Don't build the model. Simulation may fail to start
        --simulate-model=[ON, OFF]
                                Shall the simulation start automatically? [default: ON]
        --output-path=PATH      Alternative output directory, overriding BlueConfigs
        --keep-build            Keep coreneuron intermediate data. Otherwise deleted at the end
        --modelbuilding-steps=<number>
                                Set the number of ModelBuildingSteps for the CoreNeuron sim
        --experimental-stims    Shall use only Python stimuli? [default: False]
        --lb-mode=[RoundRobin, WholeCell, MultiSplit]
                                The Load Balance mode.
                                - RoundRobin: Disable load balancing. Good for quick simulations
                                - WholeCell: Does a first pass to compute load balancing and
                                    redistributes cells so that CPU load is similar among ranks
                                - MultiSplit: Allows splitting cells into pieces for distribution.
                                    WARNING: This mode is incompatible with CoreNeuron
        --save=<PATH>           Path to create a save point to enable resume.
        --save-time=<TIME>      The simulation time [ms] to save the state. (Default: At the end)
        --restore=<PATH>        Restore and resume simulation from a save point on disk
        --dump-cell-state=<GID> Dump cell state debug files on start, save-restore and at the end
        --enable-shm=[ON, OFF]  Enables the use of /dev/shm for coreneuron_input [default: ON]
        --model-stats           Show model stats in CoreNEURON simulations [default: False]
        --dry-run               Dry-run simulation to estimate memory usage [default: False]
    """
    options = docopt_sanitize(docopt(neurodamus.__doc__, args))
    config_file = options.pop("ConfigFile")
    log_level = _pop_log_level(options)

    if not os.path.isfile(config_file):
        logging.error("Config file not found: %s", config_file)
        return 1

    # Shall replace process with special? Don't if is special or already replaced
    if not sys.argv[0].endswith("special") and not os.environ.get("neurodamus_special"):
        _attempt_launch_special(config_file)

    # Warning control before starting the process
    _filter_warnings()

    try:
        Neurodamus(config_file, True, logging_level=log_level, **options).run()
    except ConfigurationError as e:  # Common, only show error in Rank 0
        if MPI._rank == 0:           # Use _rank so that we avoid init
            logging.error(str(e))
        return 1
    except OtherRankError:
        return 1  # no need for _mpi_abort, error is being handled by all ranks
    except Exception:
        show_exception_abort("Unhandled Exception. Terminating...", sys.exc_info())
        return 1
    return 0


def hocify(args=None):
    """hocify

    Usage:
        hocify <MorphologyPath> [options]
        hocify --help

    Options:
        -v --verbose            Increase verbosity level.
        --nframe=<number>       NEURON_NFRAME value [default: 1000].
        --output-dir=<PATH>     Output directory for hoc files.
    """
    options = docopt_sanitize(docopt(hocify.__doc__, args))
    morph_path = abspath(options.pop("MorphologyPath"))
    log_level = _pop_log_level(options)
    neuron_nframe = options.pop("nframe")
    options.pop("help")  # never pass to the library

    # first check if it is a file
    if os.path.isfile(morph_path):
        os.environ['NEURON_NFRAME'] = str(neuron_nframe)
        # the dest file is the same as the morph file but with .hoc extension
        dest_file = Path(morph_path).with_suffix('.hoc')

        print("Hocifying {} -> {}".format(morph_path, dest_file))
        result = hocify_process_file((morph_path, str(dest_file)))
        if isinstance(result, Exception):
            logging.critical(str(result), exc_info=True)
            return 1
        print('Done.')
        return 0

    # otherwise it is a directory, use multiprocessing
    try:
        Hocify(morph_path, neuron_nframe, log_level, **options).convert()
    except Exception as e:
        logging.critical(str(e), exc_info=True)
        return 1
    from neuron import version as nrn_version
    logging.info("Neuron version used for hocifying: " + nrn_version)
    return 0


def _pop_log_level(options):
    log_level = LogLevel.DEFAULT
    if options.pop("debug", False):
        log_level = LogLevel.DEBUG
    elif options.pop("verbose", False):
        log_level = LogLevel.VERBOSE
    if log_level >= 3:
        from pprint import pprint
        pprint(options)
    return log_level


def show_exception_abort(err_msg, exc_info):
    """Show an exception info in only one rank

    Several ranks are likely to be in sync so a simple touch wont work.
    Processes that dont see any file will register (append) their rank id
    First one is elected to print
    """
    err_file = Path(EXCEPTION_NODE_FILENAME)
    ALL_RANKS_SYNC_WINDOW = 1

    if err_file.exists():
        return 1

    with open(err_file, 'a') as f:
        f.write(str(MPI.rank) + "\n")

    with open(err_file, 'r') as f:
        line0 = open(err_file).readline().strip()
    if str(MPI.rank) == line0:
        logging.critical(err_msg, exc_info=exc_info)

    time.sleep(ALL_RANKS_SYNC_WINDOW)  # give time to the rank that logs the exception
    _mpi_abort()  # abort all ranks which have waited. Seems to help avoiding MPT stack


def _attempt_launch_special(config_file):
    import shutil
    special = shutil.which("special")
    if os.path.isfile("x86_64/special"):  # prefer locally compiled special
        special = abspath("x86_64/special")
    if special is None:
        logging.warning("special not found. Running neurodamus from Python with libnrnmech. "
                        "-> DO NOT USE WITH PRODUCTION RUNS")
        return
    neurodamus_py_root = os.environ.get("NEURODAMUS_PYTHON")
    if not neurodamus_py_root:
        logging.warning("No NEURODAMUS_PYTHON set. Running neurodamus from Python with libnrnmech. "
                        "-> DO NOT USE WITH PRODUCTION RUNS")
        return
    print("::INIT:: Special available. Replacing binary...")
    os.environ["neurodamus_special"] = "1"
    init_script = os.path.join(neurodamus_py_root, "init.py")
    os.execl(special,
             "-mpi",
             "-python", init_script,
             "--configFile=" + config_file,
             *sys.argv[2:])


def _mpi_abort():
    import ctypes
    c_api = ctypes.CDLL(None)
    c_api.MPI_Abort(0)


def _filter_warnings():
    """ Control matched warning to display once in rank 0.

    Warning 1:
    "special" binaries built with %intel build_type=Release,RelWithDebInfo flushes
    denormal results to zero, which triggers the numpy warning for subnormal in every rank.
    Reduce this type of warning displayed once in rank0.
    Note: "special" with build_type = FastDebug/Debug or calling the simulation process
       in python (built with gcc) does not have such flush-to-zero warning.
    """
    import warnings
    if MPI.rank == 0:
        action = "once"
    else:
        action = "ignore"
    warnings.filterwarnings(action=action,
                            message="The value of the smallest subnormal for .* type is zero.",
                            category=UserWarning, module="numpy")



#!/usr/bin/env python
"""
A tutorial based on the Neuron documentation at
https://neuron.yale.edu/neuron/static/docs/neuronpython/firststeps.html#step-2-create-a-cell
"""
from neurodamus.core import Cell, Neuron
from neurodamus.core import mechanisms
from neurodamus.core import synapses

c = Cell.Builder.add_soma(100).create()
mechanisms.PAS().apply(c.soma)

syn = synapses.AlphaSynapse(c.soma(0.5), onset=20, gmax=1)

Neuron.run_sim(40, c.soma(0.5)).plot()



from __future__ import print_function
from contextlib import contextmanager
import time as _time


@contextmanager
def time():
    start = _time.time()
    yield
    print("The function took %g secs." % (_time.time() - start))


def timeit():
    tot_time = .0
    loop = 0

    start = _time.time()
    yield 0
    stop = _time.time()
    best_time = stop - start
    tot_time = best_time

    while tot_time < 4:
        loop += 1
        start = stop
        yield loop
        stop = _time.time()
        tdiff = stop - start
        best_time = min(best_time, tdiff)
        tot_time += tdiff

    print("Best run: %.3f ms. [%d loops]" % (best_time * 1000, loop + 1))



"""
Classes to represent a task's progress in the form of a progress bar.

Here is some basic usage with the default options:

    >>> from progressbar import ProgressBar
    >>> p = ProgressBar()
    >>> print p
    [>............] 0%
    >>> p + 1
    >>> print p
    [=>...........] 10%
    >>> p + 9
    >>> print p
    [============>] 0%

And here another example with different options:

    >>> from progressbar import ProgressBar
    >>> custom_options = {
    ...     'end': 100,
    ...     'width': 20,
    ...     'fill': '#',
    ...     'format': '%(progress)s%% [%(fill)s%(blank)s]'
    ... }
    >>> p = ProgressBar(**custom_options)
    >>> print p
    0% [....................]
    >>> p + 5
    >>> print p
    5% [#...................]
    >>> p + 9
    >>> print p
    100% [####################]
"""
from __future__ import print_function
import sys
import time
from itertools import islice


class Progress(object):
    """Progress class holds the progress information.

    It can be queried for the current progress, and overloads __repr__ for a simple display.
    It can be managed manually, via '+=' and -= operators, or automatically by consuming
    iterables.
    """

    def __init__(self, end, start=0, **kw):
        """ Creates a progress bar

        Args:
            end:   State in which the progress has terminated. False for unknown (-> spinner)
            start: State from which start the progress. For example, if start is
                   5 and the end is 10, the progress of this state is 50%
        """
        if start < 0 or (end is not False and start > end):
            raise ValueError("Invalid Start value. Must be a non-negative smaller than end")
        self._start = start
        self._end = end
        self._init_time = time.time()
        self.reset()

    def __iadd__(self, increment):
        self.progress += increment
        if self._end is not False:
            self.progress = min(self._end, self.progress)
        return self

    def __isub__(self, decrement):
        self.progress = max(self._start, self.progress - decrement)
        return self

    @property
    def completion_ratio(self):
        try:
            return float(self.progress - self._start) / self._end
        except ZeroDivisionError:
            if self._end is False:
                return False
        return 1.0

    def __repr__(self):
        total = "N.A." if self._end is False else str(self._end)
        return "<Progress: %d/%s>" % (self.progress, total)

    def reset(self):
        """Resets the current progress to the start point"""
        self.progress = self._start

    def _set_progress(self, val):
        self._progress = val

    progress = property(lambda self: self._progress, _set_progress)

    @property
    def time_taken(self):
        return time.time() - self._init_time

    def __call__(self, iterable, end=None, start=0):
        for elem in islice(iterable, start, end):
            yield elem
            self.__iadd__(1)

    # Helpers to monitor/show progress while consuming an iterable
    # ------------------------------------------------------------
    @classmethod
    def iter(cls, iterable, end=None, start=0, **kw):
        """Consumes (a slice of) an iterable.

        Args:
            iterable: the iterable to consume and monitor progress
            end: The end index. Alternatively None will automatically detect size,
                 while False instructs to not compute size -> spinner
            start: in which position to start iterating
        """
        if end is None:
            try: end = len(iterable)
            except TypeError: end = False
        # __call__ 'end' cant be False -> None
        return cls(end, start, **kw)(iterable, end or None, start)

    @classmethod
    def itervalues(cls, iterable, **kw):
        return cls.iter(iterable.values(), len(iterable), **kw)

    @classmethod
    def iteritems(cls, iterable, **kw):
        return cls.iter(iterable.items(), len(iterable), **kw)


class ProgressBar(Progress):
    """ProgressBar implements a fully visual text-based representation of a progress.
    and may be any file-object to which send the progress status.
    """
    _no_tty_bar = "-------20%-------40%-------60%-------80%------100%"  # len 50

    def __init__(self, end, start=0, width=60, fill='=', blank='.', stream=sys.stdout,
                 clear=None, fmt='[%(fill)s>%(blank)s] %(progress)s', tty_bar=None, name=''):
        """
        Args:
            end:   State in which the progress has terminated. False for unknown (-> spinner)
            start: State from which start the progress. For example, if start is
                   5 and the end is 10, the progress of this state is 50%
            width: bar length
            fill:  String to use for "filled" used to represent the progress
            blank: String to use for "filled" used to represent remaining space.
            stream: the destination stream (default: stdout),
            clear: whether to clear the current line or keep time info (and add '\n')
            fmt: Bar format string
            tty_bar: Controls whether the bar should be enhanced for text terminals.
                Default: None (auto-detect), False, True
        """
        self._tty_mode = ((hasattr(stream, 'isatty') and stream.isatty())
                          if tty_bar is None else tty_bar)
        self._width = width if self._tty_mode else 50
        self._fill = fill
        self._blank = blank
        self._format = fmt + " "
        self._stream = stream
        self._clear = self._tty_mode if clear is None else clear
        self._prev_bar_len = None
        self._name = name
        super(ProgressBar, self).__init__(end, start)

    def _bar_len_progress(self):
        if self._end is not False:
            ratio = self.completion_ratio
            return int(self._width * ratio), "{}%".format(int(ratio * 100))
        else:
            return self.progress % self._width, str(self.progress)

    def __str__(self):
        bar_len, progress = self._bar_len_progress()
        fill = self._fill * bar_len
        blank = self._blank * (self._width - bar_len)
        return self._name + self._format % {'fill': fill, 'blank': blank, 'progress': progress}

    def show_progress(self):
        if self._end == 0:
            return
        if self._tty_mode:
            self._stream.write('\r' + str(self))
        else:
            self._show_incremental_bar()
        self._stream.flush()

    def _show_incremental_bar(self):
        bar_len, _ = self._bar_len_progress()
        if self._prev_bar_len is None or bar_len < self._prev_bar_len:
            # We need to produce a new bar.
            self._stream.write("\r{}|".format(self._name))
            self._prev_bar_len = 0

        if bar_len > self._prev_bar_len:
            self._stream.write(self._no_tty_bar[self._prev_bar_len:bar_len])
            self._prev_bar_len = bar_len

    def __del__(self):
        if not self._tty_mode:
            # Complete progressbar
            self._show_incremental_bar()

        if self._clear:
            self._stream.write("\r{}\r".format(" " * (self._width + 8 + len(self._name))))
        else:
            # Output a nice stat about time taken
            self._stream.write("| [Time taken: {:.2f} s]\n".format(self.time_taken))

    def _set_progress(self, val):
        Progress._set_progress(self, val)
        self.show_progress()

    progress = property(lambda self: self._progress, _set_progress)


# ------------------------
# QUICK TESTING
# ------------------------

if __name__ == '__main__':
    p = ProgressBar(100, width=80)

    while p.progress < 100:
        time.sleep(0.1)
        p += 5

    for i in range(80, -1, -20):
        time.sleep(0.5)
        p.progress = i

    del p

    # progressbars can also be used as a consumer-generators to monitor loop progress
    l1 = range(0, 100, 10)

    for j in ProgressBar.iter(l1):
        # Do something with j
        time.sleep(0.2)

    # Instances can also consume, be reused, and work with sub-selection
    # It is the user responsability to initialize with the right number of iterations
    bar = ProgressBar(15)
    l2 = range(100, 200, 10)
    for x in bar(l1):
        time.sleep(0.2)
    l2 = [x for x in bar(l2, 5)]
    assert l2 == [100, 110, 120, 130, 140]

    # Alternativelly, if unknown, use False (creates a spinner)
    bar = ProgressBar(False)
    for x in bar(l1):
        time.sleep(0.2)
    for y in bar(range(60)):
        time.sleep(0.04)



import itertools
import logging
import os.path
from abc import ABCMeta, abstractmethod
from functools import lru_cache
from typing import List

import libsonata
import numpy

from .core import MPI, NeurodamusCore as Nd
from .core.configuration import ConfigurationError, SimConfig, GlobalConfig, find_input_file
from .core.nodeset import _NodeSetBase, NodeSet, SelectionNodeSet
from .utils import compat
from .utils.logging import log_verbose


class TargetError(Exception):
    """A Exception class specific to data error with targets and nodesets"""


class TargetSpec:
    """Definition of a new-style target, accounting for multipopulation"""

    GLOBAL_TARGET_NAME = "_ALL_"

    def __init__(self, target_name):
        """Initialize a target specification

        Args:
            target_name: the target name. For specifying a population use
                the format ``population:target_name``
        """
        if target_name and ":" in target_name:
            self.population, self.name = target_name.split(":")
        else:
            self.name = target_name
            self.population = None
        if self.name == "":
            self.name = None

    def __str__(self):
        return (
            (self.name or "")
            if self.population is None
            else "{}:{}".format(self.population, self.name or "")
        )

    def __repr__(self):
        return "<TargetSpec: " + str(self) + ">"

    @property
    def simple_name(self):
        if self.name is None and self.population is None:
            return self.GLOBAL_TARGET_NAME
        return self.__str__().replace(":", "_")

    @property
    def is_full(self):
        return (self.name or "Mosaic") == "Mosaic"

    def matches(self, pop, target_name):
        """Check if it matches a given target. Mosaic and (empty) are equivalent"""
        return pop == self.population and (target_name or "Mosaic") == (self.name or "Mosaic")

    def disjoint_populations(self, other):
        # When a population is None we cannot draw conclusions
        #  - In Sonata there's no filtering and target may have multiple
        #  - In BCs it's the base population, but the specified one may be the same
        if self.population is None or other.population is None:
            return False
        # We are only sure if both are specified and different
        return self.population != other.population

    def overlap_byname(self, other):
        return self.is_full or other.is_full or self.name == other.name

    def overlap(self, other):
        """Are these target specs bound to overlap?
        If not, they still might be overlap, but Target gids need to be inspected
        """
        return self.population == other.population and self.overlap_byname(other)

    def __eq__(self, other):
        return self.matches(other.population, other.name)


class TargetManager:

    def __init__(self, run_conf):
        """
        Initializes a new TargetManager
        """
        self._run_conf = run_conf
        self.parser = Nd.TargetParser()
        self._has_hoc_targets = False
        self.hoc = None  # The hoc level target manager
        self._targets = {}
        self._nodeset_reader = self._init_nodesets(run_conf)
        if MPI.rank == 0:
            self.parser.isVerbose = 1
        # A list of the local node sets
        self.local_nodes = []

    @classmethod
    def _init_nodesets(cls, run_conf):
        config_nodeset_file = run_conf.get("config_node_sets_file", None)
        simulation_nodesets_file = run_conf.get("node_sets_file")
        if not simulation_nodesets_file and "TargetFile" in run_conf:
            target_file = run_conf["TargetFile"]
            if target_file.endswith(".json"):
                simulation_nodesets_file = target_file
        return (config_nodeset_file or simulation_nodesets_file) and \
               NodeSetReader(config_nodeset_file, simulation_nodesets_file)

    def load_targets(self, circuit):
        """Provided that the circuit location is known and whether a user.target file has been
        specified, load any target files via a TargetParser.
        Note that these will be moved into a TargetManager after the cells have been distributed,
        instantiated, and potentially split.
        """
        def _is_sonata_file(file_name):
            if file_name.endswith(".h5"):
                return True
            return False
        if circuit.CircuitPath:
            self._try_open_start_target(circuit)

        nodes_file = circuit.get("CellLibraryFile")
        if nodes_file and _is_sonata_file(nodes_file) and self._nodeset_reader:
            self._nodeset_reader.register_node_file(find_input_file(nodes_file))

    def _try_open_start_target(self, circuit):
        start_target_file = os.path.join(circuit.CircuitPath, "start.target")
        if not os.path.isfile(start_target_file):
            log_verbose("Circuit %s start.target not available! Skipping", circuit._name)
        else:
            self.parser.open(start_target_file, False)
            self._has_hoc_targets = True

    def load_user_target(self):
        # Old target files. Notice new targets with same should not happen
        target_file = self._run_conf.get("TargetFile")
        if not target_file or target_file.endswith(".json"):  # allow any ext, except nodesets
            return
        user_target = find_input_file(target_file)
        self.parser.open(user_target, True)
        self._has_hoc_targets = True
        if MPI.rank == 0:
            logging.info(" => Loaded %d targets", self.parser.targetList.count())
            if GlobalConfig.verbosity >= 3:
                self.parser.printCellCounts()

    @classmethod
    def create_global_target(cls):
        # In blueconfig mode the _ALL_ target refers to base single population)
        if not SimConfig.is_sonata_config:
            return _HocTarget(TargetSpec.GLOBAL_TARGET_NAME, None)  # None population -> generic
        return NodesetTarget(TargetSpec.GLOBAL_TARGET_NAME, [])

    def register_target(self, target):
        self._targets[target.name] = target
        hoc_target = target.get_hoc_target()
        if hoc_target:
            self.parser.updateTargetList(target)

    def register_local_nodes(self, local_nodes):
        """Registers the local nodes so that targets can be scoped to current rank"""
        self.local_nodes.append(local_nodes)
        self.parser.updateTargets(local_nodes.final_gids(), 1)

    def clear_simulation_data(self):
        self.local_nodes.clear()
        self.parser.updateTargets(Nd.Vector(), 0)
        self.init_hoc_manager(None)  # Init/release cell manager

    def get_target(self, target_spec: TargetSpec, target_pop=None):
        """Retrieves a target from any .target file or Sonata nodeset files.

        Targets are generic groups of cells not necessarily restricted to a population.
        When retrieved from the source files they can be cached.
        Targets retrieved from Sonata nodesets keep a reference to all Sonata
        node datasets and can be asked for a sub-target of a specific population.
        """
        if not isinstance(target_spec, TargetSpec):
            target_spec = TargetSpec(target_spec)
        if target_pop:
            target_spec.population = target_pop
        target_name = target_spec.name or TargetSpec.GLOBAL_TARGET_NAME
        target_pop = target_spec.population

        def get_concrete_target(target):
            """Get a more specific target, depending on specified population prefix"""
            target.update_local_nodes(self.local_nodes)
            return target if target_pop is None else target.make_subtarget(target_pop)

        # Check cached
        if target_name in self._targets:
            target = self._targets[target_name]
            return get_concrete_target(target)

        # Check if we can get a Nodeset
        target = self._nodeset_reader and self._nodeset_reader.read_nodeset(target_name)
        if target is not None:
            log_verbose("Retrieved `%s` from Sonata nodeset", target_spec)
            self.register_target(target)
            return get_concrete_target(target)

        if self._has_hoc_targets:
            if self.hoc is not None:
                log_verbose("Retrieved `%s` from Hoc TargetManager", target_spec)
                hoc_target = self.hoc.getTarget(target_name)
            else:
                log_verbose("Retrieved `%s` from the Hoc TargetParser", target_spec)
                hoc_target = self.parser.getTarget(target_name)
            target = _HocTarget(target_name, hoc_target)
            self._targets[target_name] = target
            return get_concrete_target(target)

        raise ConfigurationError(
            "Target {} can't be loaded. Check target sources".format(target_name)
        )

    def init_hoc_manager(self, cell_manager):
        # give a TargetManager the TargetParser's completed targetList
        self.hoc = Nd.TargetManager(self.parser.targetList, cell_manager)

    def get_target_points(self, target, cell_manager, **kw):
        """Helper to retrieve the points of a target.
        If target is a cell then uses compartmentCast to obtain its points.
        Otherwise returns the result of calling getPointList directly on the target.

        Args:
            target: The target name or object (faster)
            manager: The cell manager to access gids and metype infos

        Returns: The target list of points
        """
        if isinstance(target, TargetSpec):
            target = self.get_target(target)
        return target.getPointList(cell_manager, **kw)

    @lru_cache()
    def intersecting(self, target1, target2):
        """Checks whether two targets intersect"""
        target1_spec = TargetSpec(target1)
        target2_spec = TargetSpec(target2)
        if target1_spec.disjoint_populations(target2_spec):
            return False
        if target1_spec.overlap(target2_spec):
            return True

        # Couldn't get any conclusion from bare target spec
        # Obtain the targets to analyze
        t1, t2 = self.get_target(target1_spec), self.get_target(target2_spec)

        # Check for Sonata nodesets, they might have the same population and overlap
        if set(t1.populations) == set(t2.populations):
            if target1_spec.overlap_byname(target2_spec):
                return True

        # TODO: Investigate this might yield different results depending on the rank.
        return t1.intersects(t2)  # Otherwise go with full gid intersection

    def pathways_overlap(self, conn1, conn2, equal_only=False):
        src1, dst1 = conn1["Source"], conn1["Destination"]
        src2, dst2 = conn2["Source"], conn2["Destination"]
        if equal_only:
            return TargetSpec(src1) == TargetSpec(src2) and TargetSpec(dst1) == TargetSpec(dst2)
        return self.intersecting(src1, src2) and self.intersecting(dst1, dst2)

    def __getattr__(self, item):
        logging.debug("Compat interface to TargetManager::" + item)
        return getattr(self.hoc, item)


class NodeSetReader:
    """
    Implements reading Sonata Nodesets
    """

    def __init__(self, config_nodeset_file, simulation_nodesets_file):
        def _load_nodesets_from_file(nodeset_file):
            if not nodeset_file:
                return libsonata.NodeSets("{}")
            return libsonata.NodeSets.from_file(nodeset_file)
        self._population_stores = {}
        self.nodesets = _load_nodesets_from_file(config_nodeset_file)
        simulation_nodesets = _load_nodesets_from_file(simulation_nodesets_file)
        duplicate_nodesets = self.nodesets.update(simulation_nodesets)
        if duplicate_nodesets:
            logging.warning("Some node set rules were replaced from %s", simulation_nodesets_file)

    def register_node_file(self, node_file):
        storage = libsonata.NodeStorage(node_file)
        for pop_name in storage.population_names:
            self._population_stores[pop_name] = storage

    def __contains__(self, nodeset_name):
        return nodeset_name in self.nodesets.names

    @property
    def names(self):
        return self.nodesets.names

    def read_nodeset(self, nodeset_name: str):
        """Build node sets capable of offsetting.
        The empty population has a special meaning in Sonata, it matches
        all populations in simulation
        """
        if nodeset_name not in self.nodesets.names:
            return None

        def _get_nodeset(pop_name):
            storage = self._population_stores.get(pop_name)
            population = storage.open_population(pop_name)
            # Create NodeSet object with 1-based gids
            try:
                node_selection = self.nodesets.materialize(nodeset_name, population)
            except libsonata.SonataError as e:
                logging.warning("SonataError for nodeset %s from population \"%s\" : %s, skip"
                                % (nodeset_name, pop_name, str(e)))
                return None
            if node_selection:
                logging.debug("Nodeset %s: Appending gis from %s", nodeset_name, pop_name)
                ns = SelectionNodeSet(node_selection)
                ns.register_global(pop_name)
                return ns
            return None

        nodesets = (_get_nodeset(pop_name) for pop_name in self._population_stores)
        nodesets = [ns for ns in nodesets if ns]
        return NodesetTarget(nodeset_name, nodesets)


class _TargetInterface(metaclass=ABCMeta):
    """
    Methods that target/target wrappers should implement
    """

    @abstractmethod
    def gid_count(self):
        return NotImplemented

    @abstractmethod
    def get_gids(self):
        return NotImplemented

    @abstractmethod
    def get_raw_gids(self):
        return NotImplemented

    @abstractmethod
    def __contains__(self, final_gid):
        """
        Checks if a gid (with offset) is present in this target.
        All gids are taken into consideration, not only this ranks.
        """
        return NotImplemented

    @abstractmethod
    def make_subtarget(self, pop_name):
        return NotImplemented

    @abstractmethod
    def is_void(self):
        return NotImplemented

    @abstractmethod
    def get_hoc_target(self):
        return NotImplemented

    @abstractmethod
    def append_nodeset(self, nodeset: NodeSet):
        """Add a nodeset to the current target"""
        return NotImplemented

    def contains(self, items, raw_gids=False):
        """Return a bool or an array of bool's whether the elements are contained
        """
        # Shortcut for empty target. Algorithm below would fail
        if not self.gid_count():
            return ([False] * len(items)) if hasattr(items, "__len__") else False

        gids = self.get_raw_gids() if raw_gids else self.get_gids()
        pos = numpy.searchsorted(gids, items)
        if pos.ndim == 0:
            return pos < gids.size and gids[pos] == items
        else:
            pos[pos == len(gids)] = 0  # arbitrarily change to valid pos
            return gids[pos] == items

    def intersects(self, other):
        """ Check if two targets intersect. At least one common population has to intersect
        """
        if self.population_names.isdisjoint(other.population_names):
            return False

        other_pops = other.populations  # may be created on the fly
        # We loop over one target populations and check the other existence and intersection
        for pop, nodeset in self.populations.items():
            if pop not in other_pops:
                continue
            if nodeset.intersects(other_pops[pop]):
                return True
        return False

    @abstractmethod
    def generate_subtargets(self, n_parts):
        return NotImplemented

    def update_local_nodes(self, _local_nodes):
        """Allows setting the local gids"""
        pass


class _HocTargetInterface:
    """
    Interface of Hoc targets to be respected when we want to use objects
    in place of hoc targets.

    This interface provides a default implementation suitable for Nodeset targets
    where it will be primarily used
    """

    def isCellTarget(self, **kw):
        section_type = kw.get("sections", "soma")
        compartment_type = kw.get("compartments", "center" if section_type == "soma" else "all")
        return section_type == "soma" and compartment_type == "center"

    def isCompartmentTarget(self, *_):
        return 0

    def isSectionTarget(self, *_):
        return 0

    def isSynapseTarget(self, *_):
        return 0

    def getCellCount(self):
        return self.gid_count()

    def completegids(self):
        return compat.hoc_vector(self.get_raw_gids())

    def gids(self):
        return compat.hoc_vector(self.get_local_gids())

    @abstractmethod
    def getPointList(self, _cell_manager, **_kw):
        return NotImplemented

    def set_offset(self, *_):
        pass  # Only hoc targets require manually setting offsets

    def get_offset(self, *_):
        pass  # nodeset targets can span multiple multipopulation -> multiple offsets


class NodesetTarget(_TargetInterface, _HocTargetInterface):
    def __init__(self, name, nodesets: List[_NodeSetBase], local_nodes=None, **_kw):
        self.name = name
        self.nodesets = nodesets
        self.local_nodes = local_nodes

    def gid_count(self):
        return sum(len(ns) for ns in self.nodesets)

    def get_gids(self):
        """ Retrieve the final gids of the nodeset target """
        if not self.nodesets:
            logging.warning("Nodeset '%s' can't be materialized. No node populations", self.name)
            return numpy.array([])
        nodesets = sorted(self.nodesets, key=lambda n: n.offset)  # Get gids ascending
        gids = nodesets[0].final_gids()
        for extra_nodes in nodesets[1:]:
            gids = numpy.append(gids, extra_nodes.final_gids())
        return gids

    def get_raw_gids(self):
        """ Retrieve the raw gids of the nodeset target """
        if not self.nodesets:
            logging.warning("Nodeset '%s' can't be materialized. No node populations", self.name)
            return []
        if len(self.nodesets) > 1:
            raise TargetError("Can not get raw gids for Nodeset target with multiple populations.")
        return numpy.array(self.nodesets[0].raw_gids())

    def __contains__(self, gid):
        """ Determine if a given gid is included in the gid list for this target
        regardless of which cpu. Offsetting is taken into account
        """
        return self.contains(gid)

    def append_nodeset(self, nodeset: NodeSet):
        self.nodesets.append(nodeset)

    @property
    def population_names(self):
        return {ns.population_name for ns in self.nodesets}

    @property
    def populations(self):
        return {ns.population_name: ns for ns in self.nodesets}

    def make_subtarget(self, pop_name):
        """A nodeset subtarget contains only one given population
        """
        nodesets = [ns for ns in self.nodesets if ns.population_name == pop_name]
        local_nodes = [n for n in self.local_nodes if n.population_name == pop_name]
        return NodesetTarget(f"{self.name}#{pop_name}", nodesets, local_nodes)

    def is_void(self):
        return len(self.nodesets) == 0

    def get_hoc_target(self):
        return self  # impersonate a hoc target

    def update_local_nodes(self, local_nodes):
        self.local_nodes = local_nodes

    def get_local_gids(self, raw_gids=False):
        """Return the list of target gids in this rank (with offset)
        """
        assert self.local_nodes, "Local nodes not set"

        def pop_gid_intersect(nodeset: _NodeSetBase, raw_gids=False):
            for local_ns in self.local_nodes:
                if local_ns.population_name == nodeset.population_name:
                    return nodeset.intersection(local_ns, raw_gids)
            return []

        if raw_gids:
            assert len(self.nodesets) == 1, "Multiple populations when asking for raw gids"
            return pop_gid_intersect(self.nodesets[0], raw_gids=True)

        # If target is named Mosaic, basically we don't filter and use local_gids
        if self.name == "Mosaic" or self.name.startswith("Mosaic#"):
            gids_groups = tuple(n.final_gids() for n in self.local_nodes)
        else:
            gids_groups = tuple(pop_gid_intersect(ns) for ns in self.nodesets)

        return numpy.concatenate(gids_groups) if gids_groups else numpy.empty(0)

    def getPointList(self, cell_manager, **kw):
        """ Retrieve a TPointList containing compartments (based on section type and
        compartment type) of any local cells on the cpu.
        Args:
            cell_manager: a cell manager or global cell manager
            sections: section type, such as "soma", "axon", "dend", "apic" and "all",
                      default = "soma"
            compartments: compartment type, such as "center" and "all",
                          default = "center" for "soma", default = "all" for others
        Returns:
            list of TPointList containing the compartment position and retrieved section references
        """
        section_type = kw.get("sections") or "soma"
        compartment_type = kw.get("compartments") or ("center" if section_type == "soma" else "all")
        pointList = compat.List()
        for gid in self.get_local_gids():
            point = Nd.TPointList(gid)
            cellObj = cell_manager.get_cellref(gid)
            secs = getattr(cellObj, section_type)
            for sec in secs:
                if compartment_type == "center":
                    point.append(Nd.SectionRef(sec), 0.5)
                else:
                    for seg in sec:
                        point.append(Nd.SectionRef(sec), seg.x)
            pointList.append(point)
        return pointList

    def generate_subtargets(self, n_parts):
        """generate sub NodeSetTarget per population for multi-cycle runs
        Returns:
            list of [sub_target_n_pop1, sub_target_n_pop2, ...]
        """
        if not n_parts or n_parts == 1:
            return False

        all_raw_gids = {ns.population_name: ns.final_gids() - ns.offset for ns in self.nodesets}
        from collections import defaultdict
        new_targets = defaultdict(list)
        pop_names = list(all_raw_gids.keys())

        for cycle_i in range(n_parts):
            for pop in pop_names:
                # name sub target per populaton, to be registered later
                target_name = "{}__{}_{}".format(pop, self.name, cycle_i)
                target = NodesetTarget(target_name, [NodeSet().register_global(pop)])
                new_targets[pop].append(target)

        for pop, raw_gids in all_raw_gids.items():
            target_looper = itertools.cycle(new_targets[pop])
            for gid in raw_gids:
                target = next(target_looper)
                target.nodesets[0].add_gids([gid])

        # return list of subtargets lists of all pops per cycle
        return [[targets[cycle_i] for targets in new_targets.values()]
                    for cycle_i in range(n_parts)]


class _HocTarget(_TargetInterface):
    """
    A wrapper around Hoc targets to implement _TargetInterface
    """
    GID_DTYPE = numpy.uint32

    def __init__(self, name, hoc_target, pop_name=None, *, _raw_gids=None):
        self.name = name
        self.population_name = pop_name
        self.hoc_target = hoc_target
        self.offset = 0
        self._raw_gids = _raw_gids and numpy.array(_raw_gids, dtype=self.GID_DTYPE)

    @property
    def population_names(self):
        return {self.population_name}

    @property
    def populations(self):
        return {self.population_name: NodeSet(self.get_gids())}

    def gid_count(self):
        return len(self.get_raw_gids())

    def get_gids(self):
        if not self.offset:
            return self.get_raw_gids()
        try:
            return numpy.add(self.get_raw_gids(), self.offset, dtype=self.GID_DTYPE)
        except numpy.core._exceptions.UFuncTypeError as e:
            logging.error("Type error: please use type uint32 for the array of raw gids.")
            raise e

    def get_raw_gids(self):
        if self._raw_gids is None:
            assert self.hoc_target
            self._raw_gids = self.hoc_target.completegids().as_numpy().astype(self.GID_DTYPE)
            self._raw_gids.sort()
        return self._raw_gids

    def get_hoc_target(self):
        return self.hoc_target

    def gids(self):
        """This target gids on this rank, with offset"""
        return self.hoc_target.gids()

    def get_local_gids(self):
        return self.hoc_target.gids().as_numpy().astype(self.GID_DTYPE)

    def getPointList(self, cell_manager, **kw):
        return self.hoc_target.getPointList(cell_manager)

    def make_subtarget(self, pop_name):
        if pop_name is not None:
            # Old targets have only one population. Ensure one doesn't assign more than once
            if self.name == TargetSpec.GLOBAL_TARGET_NAME:  # This target is special
                return _HocTarget(pop_name + "__ALL__", Nd.Target(self.name), pop_name)
            if self.population_name not in (None, pop_name):
                raise ConfigurationError("Target %s cannot be reassigned population %s (cur: %s)"
                                         % (self.name, pop_name, self.population_name))
            self.population_name = pop_name
        return self

    def append_nodeset(self, nodeset: NodeSet):
        # Not very common but we may want to set the nodes later (e.g. the _ALL_ target)
        if self.population_name is not None:
            logging.warning("[Compat] Skipping adding population %s to HOC target %s",
                            nodeset.population_name, self.name)
            return
        self.population_name = nodeset.population_name
        self.offset = nodeset.offset
        self._raw_gids = numpy.asarray(nodeset.raw_gids())
        hoc_gids = compat.hoc_vector(self._raw_gids)
        self.hoc_target = Nd.Target(self.name, hoc_gids, self.population_name)

    def __contains__(self, item):
        return self.hoc_target.completeContains(item)

    def is_void(self):
        return not bool(self.hoc_target)  # old targets could match with any population

    def set_offset(self, offset):
        self.hoc_target.set_offset(offset)
        self.offset = offset

    def isCellTarget(self, **kw):
        return self.hoc_target.isCellTarget()

    def generate_subtargets(self, n_parts):
        """generate sub hoc targets for multi-cycle runs
        Returns:
            list of subtargets
        """
        if not n_parts or n_parts == 1:
            return False

        allgids = self.get_gids()
        new_targets = []

        for cycle_i in range(n_parts):
            target = Nd.Target()
            target.name = "{}_{}".format(self.name, cycle_i)
            new_targets.append(_HocTarget(target.name, target, self.population_name))

        target_looper = itertools.cycle(new_targets)
        for gid in allgids:
            target = next(target_looper)
            target.hoc_target.gidMembers.append(gid)

        return new_targets

    def __getattr__(self, item):
        return getattr(self.hoc_target, item)



#!/usr/bin/env python
"""
Neuron tutorial II using the new HighLevel API. Original tutorial at
https://github.com/BlueBrain/SimulationTutorials/blob/master/CNS2017/NEURON/NEURON_morph_synapses.ipynb
"""
from neurodamus.core import Cell
from neurodamus.core import CurrentSource
from neurodamus.core import Neuron
from neurodamus.core import synapses
from neurodamus.core import mechanisms
from os import path as Path

MORPHO = Path.join(Path.dirname(__file__), "..", "tests/morphology/C060114A7.asc")


def test_tut2():
    c = Cell(0, MORPHO)
    print("The morphology {} basal, {} apical, {} axonal sections.".format(
        len(c.dendrites), len(c.apical_dendrites), len(c.axons)))

    for sec in c.all:
        sec.nseg = 1 + 2 * int(sec.L / 40)
        sec.cm = 1
        sec.Ra = 100

    hh = mechanisms.HH(gkbar=.0, gnabar=.0, el=-70)
    hh.apply(c.all)
    hh.gkbar = 0.01
    hh.gnabar = 0.2
    hh.apply(c.soma)
    hh.gnabar = 0.25
    hh.apply(c.axons)

    clamp = CurrentSource.Constant(0.5, 200, 10).attach_to(c.soma)
    Neuron.run_sim(300, c.soma, v_init=-70).plot()

    # Part 2
    clamp.detach()

    stim = synapses.NetStim(5, 5, 20, 0)

    for sect in c.dendrites:
        sr = synapses.ExpSyn(sect(0.5))
        stim.connect_to(sr, weight=0.001)

    Neuron.run_sim(80, c.soma, v_init=-70).plot()

    # Alternative
    # # --------
    # for sr in ExpSyn().create_on(*c.dendrites):
    #     stim.connect_to(sr, weight=0.001)


if __name__ == "__main__":
    test_tut2()



import logging

from .connection_manager import SynapseRuleManager
from .connection import Connection, NetConType, ReplayMode
from .core.configuration import GlobalConfig, SimConfig
from .io.synapse_reader import SynapseParameters, SonataReader
from .utils.logging import log_all


class NeuroModulationConnection(Connection):
    __slots__ = ("_neuromod_strength", "_neuromod_dtc")

    def __init__(self,
                 sgid, tgid, src_pop_id=0, dst_pop_id=0,
                 weight_factor=1.0,
                 minis_spont_rate=None,
                 configuration=None,
                 mod_override=None,
                 **kwargs):
        self._neuromod_strength = None
        self._neuromod_dtc = None
        super().__init__(sgid, tgid, src_pop_id, dst_pop_id,
                         weight_factor, minis_spont_rate, configuration, mod_override, **kwargs)

    neuromod_strength = property(
        lambda self: self._neuromod_strength,
        lambda self, val: setattr(self, '_neuromod_strength', val)
    )
    neuromod_dtc = property(
        lambda self: self._neuromod_dtc,
        lambda self, val: setattr(self, '_neuromod_dtc', val)
    )

    def finalize(self, cell, base_seed=0, *,
                 skip_disabled=False,
                 replay_mode=ReplayMode.AS_REQUIRED,
                 base_manager=None, attach_src_cell=True, **_kwargs):
        """ Override the finalize process from the base class.
            NeuroModulatory events do not create synapses but link to existing cell synapses.
            A neuromodulatory connection from projections with match to the closest cell synapse.
            A spike coming from the neuromodulatory event (SynapseReplay) will trigger the
            NET_RECEIVE of the existing synapse, with the weight (binary 1/0), neuromod_strength,
            neuromod_dtc, and nc_type NC_MODULATOR
        """
        logging.debug("Finalize neuromodulation connection")
        if skip_disabled and self._disabled:
            return 0

        self._netcons = []
        # Initialize member lists
        self._init_artificial_stims(cell, replay_mode)

        for syn_i, sec in self.sections_with_synapses:
            syn_params = self._synapse_params[syn_i]
            # We need to get all connections since we dont know the sgid
            # TODO: improve this by extracting all the relative distances only once
            base_conns = base_manager.get_connections(self.tgid)
            syn_obj = self._find_closest_cell_synapse(syn_params, base_conns)
            if syn_obj is None:
                logging.warning("No cell synapse associated to the neuromodulatory event")
                return 0
            nc = None
            # For coreneuron, create NetCon attached to the (virtual) source gid for replay
            # For neuron, create NetCon with source from replay stim
            if SimConfig.use_coreneuron:
                nc = self._pc.gid_connect(self.sgid, syn_obj)
                nc.delay = syn_params.delay
                self._netcons.append(nc)
            elif self._replay is not None:
                nc = self._replay.create_on(self, sec, syn_obj, syn_params)
            if nc:
                nc.weight[0] = int(self.weight_factor > 0)  # weight is binary 1/0, default 1
                nc.weight[1] = self.neuromod_strength or syn_params.neuromod_strength
                nc.weight[2] = self.neuromod_dtc or syn_params.neuromod_dtc
                self.netcon_set_type(nc, syn_obj, NetConType.NC_NEUROMODULATOR)
                if GlobalConfig.debug_conn == [self.tgid]:
                    log_all(logging.DEBUG, "Neuromodulatory event on tgid: %d, " +
                            "weights: [%f, %f, %f], nc_type: %d", self.tgid,
                            nc.weight[0], nc.weight[1], nc.weight[2], NetConType.NC_NEUROMODULATOR)

        return 1

    def _find_closest_cell_synapse(self, syn_params, base_conns):
        """ Find the closest cell synapse by the location parameter
        """
        if not base_conns:
            return None
        section_i = syn_params.isec
        location_i = syn_params.location
        min_diff = 0.05
        syn_obj = None
        for base_conn in base_conns:
            for syn_j, _ in base_conn.sections_with_synapses:
                params_j = base_conn._synapse_params[syn_j]
                if params_j.isec != section_i:
                    continue
                diff = abs(params_j.location - location_i)
                if diff < min_diff:
                    syn_obj = base_conn._synapses[syn_j]
                    min_diff = diff
        return syn_obj


class ModulationConnParameters(SynapseParameters):
    # Attribute names of synapse parameters,
    # For consistancy with standard synapses, location is computed with hoc function
    # TargetManager.locationToPoint using isec, offset, ipt
    # ipt is not read from data but -1, so that locationToPoint will set location = offset .
    # weight is a placeholder for replaystim, default to 1. and overwritten by connection weight.
    _synapse_fields = ("sgid", "delay", "isec", "offset", "neuromod_strength", "neuromod_dtc",
                       "ipt", "location", "weight")


class NeuroModulationSynapseReader(SonataReader):
    Parameters = ModulationConnParameters
    custom_parameters = {"isec", "ipt", "offset", "weight"}

    def _load_params_custom(self, _populate, _read):
        super()._load_params_custom(_populate, _read)
        _populate("weight", 1)


class NeuroModulationManager(SynapseRuleManager):
    conn_factory = NeuroModulationConnection
    SynapseReader = NeuroModulationSynapseReader

    def _finalize_conns(self, tgid, conns, base_seed, sim_corenrn, **kwargs):
        """ Override the function from the base class.
            Retrieve the base synapse connections with the same tgid.
            Pass the base connections to the finalize process of superclass,
            to be processed by NeuroModulationConnection.
        """
        base_manager = next(self.cell_manager.connection_managers.values())
        return super()._finalize_conns(tgid, conns, base_seed, sim_corenrn,
                                       base_manager=base_manager,
                                       **kwargs)



"""
Module implementing interfaces to the several synapse readers (eg.: synapsetool, Hdf5Reader)
"""
import logging
import os
from abc import abstractmethod

import libsonata
import numpy as np

from ..core import NeurodamusCore as Nd, MPI
from ..utils.logging import log_verbose


def _constrained_hill(K_half, y):
    K_half_fourth = K_half**4
    y_fourth = y**4
    return (K_half_fourth + 16) / 16 * y_fourth / (K_half_fourth + y_fourth)


class _SynParametersMeta(type):
    def __init__(cls, name, bases, attrs):
        type.__init__(cls, name, bases, attrs)
        # Init public properties of the class
        assert hasattr(cls, "_synapse_fields"), "Please define _synapse_fields class attr"
        cls.dtype = np.dtype({"names": cls._synapse_fields,
                              "formats": ["f8"] * len(cls._synapse_fields)})
        cls.empty = np.recarray(0, cls.dtype)
        if not hasattr(cls, "_optional"):
            cls._optional = ()
        # Reserved fields are used to hold extra info, don't come from edge files
        if not hasattr(cls, "_reserved"):
            cls._reserved = ()

        cls.load_fields = set(cls._synapse_fields) - set(cls._reserved)

    @property
    def all_fields(cls):
        return set(cls._synapse_fields)

    def fields(cls, exclude: set = (), with_translation: dict = None):
        fields = cls.load_fields - exclude if exclude else cls.load_fields
        if with_translation:
            return [(f, with_translation.get(f, f), f in cls._optional) for f in fields]
        else:
            return [(f, f in cls._optional) for f in fields]

    def create_array(cls, length):
        return np.recarray(length, cls.dtype)


class SynapseParameters(metaclass=_SynParametersMeta):
    """Synapse parameters, internally implemented as numpy record
    """
    _synapse_fields = ("sgid", "delay", "isec", "ipt", "offset", "weight", "U", "D", "F",
                       "DTC", "synType", "nrrp", "u_hill_coefficient", "conductance_ratio",
                       "maskValue", "location")  # total: 16

    _optional = ("u_hill_coefficient", "conductance_ratio")
    _reserved = ("maskValue", "location")

    def __new__(cls, *_):
        raise NotImplementedError()

    @classmethod
    def create_array(cls, length):
        npa = np.recarray(length, cls.dtype)
        npa.conductance_ratio = -1  # set to -1 (not-set). 0 is meaningful
        npa.maskValue = -1
        npa.location = 0.5
        return npa


class SynapseReader:
    """ Synapse Readers base class.
        Factory create() will attempt to instantiate a SONATA reader, followed by SynReaderNRN.
    """
    # Data types to read
    SYNAPSES = 0
    GAP_JUNCTIONS = 1

    def __init__(self, src, conn_type, population=None, *_, **kw):
        self._conn_type = conn_type
        self._ca_concentration = kw.get("extracellular_calcium")
        self._syn_params = {}  # Parameters cache by post-gid (previously loadedMap)
        self._open_file(src, population, kw.get("verbose", False))
        # NOTE u_hill_coefficient and conductance_scale_factor are optional, BUT
        # while u_hill_coefficient can always be readif avail, conductance reader may not.
        self._uhill_property_avail = self.has_property("u_hill_coefficient")
        self._extra_fields = tuple()
        self._extra_scale_vars = []

    def preload_data(self, ids):
        pass

    def configure_override(self, mod_override):
        if not mod_override:
            return

        override_helper = mod_override + "Helper"
        Nd.load_hoc(override_helper)

        # Read attribute names with format "attr1;attr2;attr3"
        attr_names = getattr(Nd, override_helper + "_NeededAttributes", None)
        if attr_names:
            log_verbose('Reading parameters "{}" for mod override: {}'.format(
                ", ".join(attr_names.split(";")), mod_override)
            )
            self._extra_fields = tuple(attr_names.split(";"))

        # Read attribute names with format "attr1;attr2;attr3"
        attr_names = getattr(Nd, override_helper + "_UHillScaleVariables", None)
        if attr_names:
            self._extra_scale_vars = attr_names.split(";")

    def get_synapse_parameters(self, gid):
        """Obtains the synapse parameters record for a given gid.
        """
        syn_params = self._syn_params.get(gid)
        if syn_params is None:
            syn_params = self._load_synapse_parameters(gid)

            # Modify parameters
            self._patch_delay_fp_inaccuracies(syn_params)
            if self._uhill_property_avail:
                self._scale_U_param(syn_params, self._ca_concentration, self._extra_scale_vars)
            self._syn_params[gid] = syn_params  # cache parameters
        return syn_params

    @abstractmethod
    def _load_synapse_parameters(self, gid):
        """The low level reading of synapses subclasses must override"""
        pass

    @staticmethod
    def _patch_delay_fp_inaccuracies(records):
        if len(records) == 0 or 'delay' not in records.dtype.names:
            return
        dt = Nd.dt
        records.delay = (records.delay / dt + 1e-5).astype('i4') * dt

    @staticmethod
    def _scale_U_param(syn_params, extra_cellular_calcium, extra_scale_vars):
        if len(syn_params) == 0:
            return
        if extra_cellular_calcium is None:
            return

        scale_factors = _constrained_hill(syn_params.u_hill_coefficient,
                                          extra_cellular_calcium)
        syn_params.U *= scale_factors

        for scale_var in extra_scale_vars:
            syn_params[scale_var] *= scale_factors

    @abstractmethod
    def _open_file(self, src, population, verbose=False):
        """Initializes the reader, opens the synapse file
        """

    @abstractmethod
    def has_nrrp(self):
        """Checks whether source data has the nrrp field.
        """

    @abstractmethod
    def has_property(self, field_name):
        """Checks whether source data has the given additional field.
        """

    @staticmethod
    def _get_sonata_circuit(path):
        """Returns a SONATA edge file in path if present
        """
        if path.endswith(".h5"):
            import h5py
            f = h5py.File(path, 'r')
            if "edges" in f:
                return path
        return None

    @classmethod
    def create(cls, syn_src, conn_type=SYNAPSES, population=None, *args, **kw):
        """Instantiates a synapse reader, giving preference to SonataReader
        """
        # If create() called from this class then FACTORY. Otherwise instantiate directly
        if cls is not SynapseReader:
            return cls(syn_src, conn_type, population, *args, **kw)

        kw["verbose"] = (MPI.rank == 0)
        if fn := cls._get_sonata_circuit(syn_src):
            log_verbose("[SynReader] Using SonataReader.")
            return SonataReader(fn, conn_type, population, **kw)

        # We dropped syntool (july-2023). If not SONATA try with the old nrn reader
        if os.path.isdir(syn_src) or syn_src.endswith(".h5"):
            logging.info("[SynReader] Attempting legacy hdf5 reader.")
            return SynReaderNRN(syn_src, conn_type, None, *args, **kw)

        raise FormatNotSupported(f"File: {syn_src}. Please provide SONATA edges")


class SonataReader(SynapseReader):
    """Reader for SONATA edge files.

    Uses libsonata directly and contains a bunch of workarounds to accomodate files
    created in the transition to SONATA. Also translates all GIDs from 0-based as on disk
    to the 1-based convention in Neurodamus.

    Will read each attribute for multiple GIDs at once and cache read data in a columnar
    fashion.

    FIXME Remove the caching at the np.recarray level.
    """

    SYNAPSE_INDEX_NAMES = ("synapse_index",)
    LOOKUP_BY_TARGET_IDS = True  # False to lookup by Source Ids
    Parameters = SynapseParameters  # By default we load synapses

    custom_parameters = {"isec", "ipt", "offset"}
    """Custom parameters are skipped from direct loading and trigger _load_params_custom()"""

    parameter_mapping = {
        "weight": "conductance",
        "U": "u_syn",
        "D": "depression_time",
        "F": "facilitation_time",
        "DTC": "decay_time",
        "synType": "syn_type_id",
        "nrrp": "n_rrp_vesicles",
        "conductance_ratio": "conductance_scale_factor",
    }

    def _open_file(self, src, population, _):
        try:
            from mpi4py import MPI
            hdf5_reader = libsonata.make_collective_reader(MPI.COMM_WORLD, False, True)
        except ModuleNotFoundError:
            hdf5_reader = libsonata.Hdf5Reader()

        storage = libsonata.EdgeStorage(src, hdf5_reader=hdf5_reader)
        if not population:
            assert len(storage.population_names) == 1
            population = next(iter(storage.population_names))
        self._population = storage.open_population(population)
        # _data is a cache which stores all the fields for each gid
        # E.g. {1: {"sgid": property_numpy}}
        self._data = {}

    def has_nrrp(self):
        """This field is required in SONATA."""
        return True

    def has_property(self, field_name):
        if field_name in self.SYNAPSE_INDEX_NAMES:
            return True
        return field_name in self._population.attribute_names

    def get_property(self, gid, field_name):
        """Retrieves a full pre-loaded property given a gid and the property name.
        """
        return self._data[gid][field_name]

    def preload_data(self, ids):
        """Preload SONATA fields for the specified IDs"""
        compute_fields = set(("sgid", "tgid") + self.SYNAPSE_INDEX_NAMES)
        needed_ids = sorted(set(ids) - set(self._data.keys()))

        def get_edge_and_lookup_gids(needed_ids):
            gids_0based = np.array(needed_ids, dtype="int64") - 1
            if self.LOOKUP_BY_TARGET_IDS:
                edge_ids = self._population.afferent_edges(gids_0based)
                return edge_ids, self._population.target_nodes(edge_ids) + 1
            else:
                edge_ids = self._population.efferent_edges(gids_0based)
                return edge_ids, self._population.source_nodes(edge_ids) + 1

        needed_edge_ids, lookup_gids = get_edge_and_lookup_gids(needed_ids)

        def _populate(field, data):
            # Populate cache. Unavailable entries are stored as a plain -1
            if data is None:
                data = -1
            for gid in needed_ids:
                existing_gid_data = self._data.setdefault(gid, {})
                existing_gid_data[field] = data if np.isscalar(data) else data[lookup_gids == gid]

        def _read(attribute, optional=False):
            if attribute in self._population.attribute_names:
                return self._population.get_attribute(attribute, needed_edge_ids)
            elif optional:
                log_verbose("Defaulting to -1.0 for attribute %s", attribute)
                return -1
            else:
                raise AttributeError(f"Missing attribute {attribute} in the SONATA edge file")

        # Populate the opposite node id
        if self.LOOKUP_BY_TARGET_IDS:
            _populate("sgid", self._population.source_nodes(needed_edge_ids) + 1)
        else:
            _populate("tgid", self._population.target_nodes(needed_edge_ids) + 1)

        # Make synapse index in the file explicit
        for name in sorted(self.SYNAPSE_INDEX_NAMES):
            _populate(name, needed_edge_ids.flatten())

        # Generic synapse parameters
        fields_load_sonata = self.Parameters.fields(exclude=self.custom_parameters | compute_fields,
                                                    with_translation=self.parameter_mapping)
        for (field, sonata_attr, is_optional) in sorted(fields_load_sonata):
            _populate(field, _read(sonata_attr, is_optional))

        if self.custom_parameters:
            self._load_params_custom(_populate, _read)

        # Extend Gids data with the additional requested fields
        # This has to work for when we call preload() a second/third time
        # so we are unsure about which gids were loaded what properties
        # We nevertheless can skip any base fields
        extra_fields = set(self._extra_fields) - (self.Parameters.all_fields | compute_fields)
        for field in sorted(extra_fields):
            now_needed_ids = sorted(set(gid for gid in ids if field not in self._data[gid]))
            if needed_ids != now_needed_ids:
                needed_ids = now_needed_ids
                needed_edge_ids, lookup_gids = get_edge_and_lookup_gids(needed_ids)
            sonata_attr = self.parameter_mapping.get(field, field)
            _populate(field, _read(sonata_attr))

    def _load_params_custom(self, _populate, _read):
        # Position of the synapse
        if self.has_property("afferent_section_id"):
            _populate("isec", _read("afferent_section_id"))
            # SONATA compliant synapse position: (section, section_fraction) takes precedence
            # over the older (section, segment, segment_offset) synapse position.
            #
            # Re-using field names for historical reason, where -1 means N/A.
            # FIXME Use dedicated fields
            if self.has_property("afferent_section_pos"):
                _populate("ipt", -1)
                _populate("offset", _read("afferent_section_pos"))
            # This was a temporary naming scheme
            # FIXME Circuits using this field should be fixed
            elif self.has_property("afferent_section_fraction"):
                logging.warning(
                    "Circuit uses non-standard compliant attribute `afferent_section_fraction`"
                )
                _populate("ipt", -1)
                _populate("offset", _read("afferent_section_fraction"))
            else:
                logging.warning(
                    "Circuit is missing standard compliant attribute `afferent_section_pos`"
                )
                _populate("ipt", _read("afferent_segment_id"))
                _populate("offset", _read("afferent_segment_offset"))
        else:
            # FIXME All this should go the way of the dodo
            logging.warning(
                "Circuit uses attribute notation using `morpho_` and is not SONATA compliant"
            )
            _populate("isec", _read("morpho_section_id_post"))
            if self.has_property("morpho_section_fraction_post"):
                _populate("ipt", -1)
                _populate("offset", _read("morpho_section_fraction_post"))
            else:
                _populate("ipt", _read("morpho_segment_id_post"))
                _populate("offset", _read("morpho_offset_segment_post"))

    def _load_synapse_parameters(self, gid):
        if gid not in self._data:
            self.preload_data([gid])

        data = self._data[gid]
        edge_count = len(next(iter(data.values())))

        if self._extra_fields:
            class CustomSynapseParameters(self.Parameters):
                _synapse_fields = self.Parameters._synapse_fields + self._extra_fields
            conn_syn_params = CustomSynapseParameters.create_array(edge_count)
        else:
            conn_syn_params = self.Parameters.create_array(edge_count)

        for name in self.Parameters.load_fields:
            conn_syn_params[name] = data[name]
        for name in self._extra_fields:
            conn_syn_params[name] = data[name]

        return conn_syn_params

    def get_counts(self, raw_ids, group_by):
        """
        Counts synapses and groups by the given field.
        """
        edge_ids = self._population.afferent_edges(raw_ids - 1)
        data = self._population.get_attribute(group_by, edge_ids)
        values, counts = np.unique(data, return_counts=True)
        return dict(zip(values, counts))


class SynReaderNRN(SynapseReader):
    """ Synapse Reader for NRN format only, using the hdf5_reader mod.
    """
    def __init__(self,
                 syn_src, conn_type, population=None,
                 n_synapse_files=None, local_gids=(),  # Specific to NRNReader
                 *_, **kw):
        if os.path.isdir(syn_src):
            filename = "nrn_gj.h5" if conn_type == self.GAP_JUNCTIONS else "nrn.h5"
            syn_src = os.path.join(syn_src, filename)
            log_verbose("Found nrn file: %s", filename)

        # Hdf5 reader doesnt do checks, failing badly (and cryptically) later
        if not os.path.isfile(syn_src) and not os.path.isfile(syn_src + ".1"):
            raise RuntimeError("NRN synapses file not found: " + syn_src)

        # Generic init now that we know the file
        self._n_synapse_files = n_synapse_files or 1  # needed during init
        SynapseReader.__init__(self, syn_src, conn_type, population, **kw)

        if self._n_synapse_files > 1:
            vec = Nd.Vector(len(local_gids))  # excg-location requires true vector
            for num in local_gids:
                vec.append(num)
            self._syn_reader.exchangeSynapseLocations(vec)

    def _open_file(self, syn_src, population, verbose=False):
        if population:
            raise RuntimeError("HDF5Reader doesn't support Populations.")
        log_verbose("Opening synapse file: %s", syn_src)
        self._syn_reader = Nd.HDF5Reader(syn_src, self._n_synapse_files)
        self.nrn_version = self._syn_reader.checkVersion()

    def has_nrrp(self):
        return self.nrn_version > 4

    def has_property(self, field_name):
        logging.warning("has_property() without SynapseReader returns always False")
        return False

    def _load_synapse_parameters(self, gid):
        reader = self._syn_reader
        cell_name = "a%d" % gid

        ret = reader.loadData(gid) if self._n_synapse_files > 1 \
            else reader.loadData(cell_name)

        if ret < 0:  # No dataset
            return SynapseParameters.empty
        nrow = int(reader.numberofrows(cell_name))
        if nrow == 0:
            return SynapseParameters.empty

        conn_syn_params = SynapseParameters.create_array(nrow)
        has_nrrp = self.has_nrrp()

        for i in range(nrow):
            params = conn_syn_params[i]
            params[0] = reader.getData(cell_name, i, 0)   # sgid
            params[1] = reader.getData(cell_name, i, 1)   # delay
            params[2] = reader.getData(cell_name, i, 2)   # isec
            params[3] = reader.getData(cell_name, i, 3)   # ipt
            params[4] = reader.getData(cell_name, i, 4)   # offset
            params[5] = reader.getData(cell_name, i, 8)   # weight
            params[6] = reader.getData(cell_name, i, 9)   # U
            params[7] = reader.getData(cell_name, i, 10)  # D
            params[8] = reader.getData(cell_name, i, 11)  # F
            params[9] = reader.getData(cell_name, i, 12)  # DTC
            params[10] = reader.getData(cell_name, i, 13)  # isynType
            if has_nrrp:
                params[11] = reader.getData(cell_name, i, 17)  # nrrp
            else:
                params[11] = -1

            # placeholder for u_hill_coefficient and conductance_ratio, not supported by HDF5Reader
            params[12] = -1
            params[13] = -1

        return conn_syn_params


class FormatNotSupported(Exception):
    """Exception thrown when the circuit requires SynapseTool and it is NOT built-in.
    """
    pass



"""
Collection of utility functions related to clearing the used memory in neurodamus-py or NEURON
"""
from collections import Counter
import ctypes
import ctypes.util
import logging
import math
import os
import json
import psutil
import multiprocessing

from ..core import MPI, NeurodamusCore as Nd, run_only_rank0

import numpy as np


def trim_memory():
    """
    malloc_trim - release free memory from the heap (back to the OS)

    * We should only run malloc_trim if we are using the default glibc memory allocator.
    When using a custom allocator such as jemalloc, this could cause unexpected behavior
    including segfaults.

    * The malloc_trim function returns 1 if memory was actually
    released back to the system, or 0 if it was not possible to
    release any memory.
    """

    if os.getenv("LD_PRELOAD") and "jemalloc" in os.getenv("LD_PRELOAD"):
        logging.warning("malloc_trim works only with the default glibc memory allocator. "
                        "Please avoid using jemalloc (for now, we skip malloc_trim).")
        logging.info("malloc_trim: not possible to release any memory.")
    else:
        try:
            path_libc = ctypes.util.find_library("c")
            libc = ctypes.CDLL(path_libc)
            memory_trimmed = libc.malloc_trim(0)
            if memory_trimmed:
                logging.info("malloc_trim: memory released back to the system.")
        except (OSError, AttributeError):
            logging.error("Unable to call malloc_trim.")
            logging.info("malloc_trim: not possible to release any memory.")


def pool_shrink():
    """
    Shrink NEURON ArrayPools to the exact size still being used by NEURON.
    This is useful when clearing all the Nodes from NEURON (cells and synapses) and the underlying
    data structures holding the data of their mechanisms need to be shrinked down since they are
    not used any more.
    This feature is enabled in NEURON after version 8.2.2a.
    See https://github.com/neuronsimulator/nrn/pull/2033 for more information.
    """
    try:
        cv = Nd.CVode()
        cv.poolshrink(1)
    except AttributeError:
        logging.warning("Cannot shrink NEURON ArrayPools. "
                        "NEURON does not support CVode().poolshrink(1)")


def free_event_queues():
    """
    Apart from the NEURON SelfEventQueue and TQItem pools being clear we also free them from any
    memory still referencing to.
    """
    try:
        cv = Nd.CVode()
        cv.free_event_queues()
    except AttributeError:
        logging.warning("Cannot clear NEURON event queues."
                        "NEURON does not support CVode().free_event_queues()")


def print_node_level_mem_usage():
    """Print statistics of the memory usage per compute node."""
    from ..core._shmutils import SHMUtil

    # The association of MPI ranks to compute nodes isn't
    # always know. If unknown, don't print anything.
    if not SHMUtil.is_node_id_known():
        return

    rss = SHMUtil.get_nodewise_rss()

    min_usage_mb = np.min(rss) / 2**20
    max_usage_mb = np.max(rss) / 2**20
    avg_usage_mb = np.mean(rss) / 2**20
    dev_usage_mb = np.std(rss) / 2**20

    logging.info(
        "Memusage (RSS) per node [MB]: Max=%.2lf, Min=%.2lf, Mean(Stdev)=%.2lf(%.2lf)",
        max_usage_mb,
        min_usage_mb,
        avg_usage_mb,
        dev_usage_mb
    )


def get_task_level_mem_usage():
    """Return statistics of the memory usage per MPI task."""
    usage_mb = get_mem_usage_kb() / 1024

    min_usage_mb = MPI.pc.allreduce(usage_mb, MPI.MIN)
    max_usage_mb = MPI.pc.allreduce(usage_mb, MPI.MAX)
    avg_usage_mb = MPI.pc.allreduce(usage_mb, MPI.SUM) / MPI.size

    dev_usage_mb = math.sqrt(MPI.pc.allreduce((usage_mb - avg_usage_mb) ** 2, MPI.SUM) / MPI.size)

    return min_usage_mb, max_usage_mb, avg_usage_mb, dev_usage_mb


def print_task_level_mem_usage():
    """Print statistics of the memory usage per MPI task."""

    min_usage_mb, max_usage_mb, avg_usage_mb, dev_usage_mb = get_task_level_mem_usage()

    logging.info(
        "Memusage (RSS) per task [MB]: Max=%.2lf, Min=%.2lf, Mean(Stdev)=%.2lf(%.2lf)",
        max_usage_mb,
        min_usage_mb,
        avg_usage_mb,
        dev_usage_mb
    )


def print_mem_usage():
    """
    Print memory usage information across all ranks.
    """

    print_node_level_mem_usage()
    print_task_level_mem_usage()


def get_mem_usage_kb():
    """
    Return memory usage information across all ranks, in KiloBytes
    """
    usage_kb = psutil.Process(os.getpid()).memory_info().rss / 1024

    return usage_kb


def pretty_printing_memory_mb(memory_mb):
    """
    A simple function that given a memory usage in MB
    returns a string with the most appropriate unit.
    """
    if memory_mb < 1024:
        return "%.2lf MB" % memory_mb
    elif memory_mb < 1024 ** 2:
        return "%.2lf GB" % (memory_mb / 1024)
    elif memory_mb < 1024 ** 3:
        return "%.2lf TB" % (memory_mb / 1024 ** 2)
    else:
        return "%.2lf PB" % (memory_mb / 1024 ** 3)


class SynapseMemoryUsage:
    ''' A small class that works as a lookup table
    for the memory used by each type of synapse.
    The values are in KB. The values cannot be set by the user.
    '''
    _synapse_memory_usage = {
        "ProbAMPANMDA": 1.7,
        "ProbGABAAB": 2.0,
        "Gap": 2.0,
        "Glue": 0.5
    }

    @classmethod
    def get_memory_usage(cls, count, synapse_type):
        return count * cls._synapse_memory_usage[synapse_type]


class DryRunStats:
    _MEMORY_USAGE_FILENAME = "cell_memory_usage.json"

    def __init__(self) -> None:
        self.metype_memory = {}
        self.metype_counts = Counter()
        self.synapse_counts = Counter()
        _, _, self.base_memory, _ = get_task_level_mem_usage()

    @run_only_rank0
    def estimate_cell_memory(self) -> float:
        from .logging import log_verbose
        memory_total = 0
        log_verbose("+{:=^81}+".format(" METype Memory Estimates (MiB) "))
        log_verbose("| {:^40s} | {:^10s} | {:^10s} | {:^10s} |".format(
            'METype', 'Mem/cell', 'N Cells', 'Mem Total'))
        log_verbose("+{:-^81}+".format(""))
        for metype, count in self.metype_counts.items():
            metype_mem = self.metype_memory[metype] / 1024
            metype_total = count * metype_mem
            memory_total += metype_total
            log_verbose("| {:<40s} | {:10.2f} | {:10.0f} | {:10.1f} |".format(
                metype, metype_mem, count, metype_total))
        log_verbose("+{:-^81}+".format(""))
        self.cell_memory_total = memory_total
        logging.info("  Total memory usage for cells: %s", pretty_printing_memory_mb(memory_total))
        return memory_total

    def add(self, other):
        self.metype_memory.update(other.metype_memory)
        self.metype_counts += other.metype_counts

    def collect_all_mpi(self):
        # We combine memory dict via update(). That means if a previous circuit computed
        # cells for the same METype (hopefully unlikely!) the last estimate prevails.
        self.metype_memory = MPI.py_reduce(self.metype_memory, {}, lambda x, y: x.update(y))
        self.metype_counts = self.metype_counts  # Cell counts is complete in every rank

    @run_only_rank0
    def export_cell_memory_usage(self):
        with open(self._MEMORY_USAGE_FILENAME, 'w') as fp:
            json.dump(self.metype_memory, fp, sort_keys=True, indent=4)

    def try_import_cell_memory_usage(self):
        if not os.path.exists(self._MEMORY_USAGE_FILENAME):
            return
        logging.info("Loading memory usage from %s...", self._MEMORY_USAGE_FILENAME)
        with open(self._MEMORY_USAGE_FILENAME, 'r') as fp:
            self.metype_memory = json.load(fp)

    def collect_display_syn_counts(self):
        master_counter = MPI.py_sum(self.synapse_counts, Counter())

        # Done with MPI. Use rank0 to display
        if MPI.rank != 0:
            return

        logging.info(" - Estimated synapse memory usage (MB):")
        from .logging import log_verbose
        inh_count = exc_count = 0
        log_verbose("+{:=^68}+".format(" Synapse Count "))
        log_verbose("| {:^40s} | {:^10s} | {:^10s} |".format("Synapse Type", "Family", "Count"))
        log_verbose("+{:-^68}+".format(""))
        for syn_type, count in sorted(master_counter.items()):
            is_inh = syn_type < 100
            syn_type_str = "INH" if is_inh else "EXC"
            log_verbose("| {:40.0f} | {:<10s} | {:10.0f} |".format(syn_type, syn_type_str, count))
            if is_inh:
                inh_count += count
            else:
                exc_count += count
        log_verbose("+{:-^68}+".format(""))

        in_mem = SynapseMemoryUsage.get_memory_usage(inh_count, "ProbGABAAB") / 1024
        ex_mem = SynapseMemoryUsage.get_memory_usage(exc_count, "ProbAMPANMDA") / 1024
        self.synapse_memory_total = in_mem + ex_mem
        logging.info("   - Inhibitory: %s", pretty_printing_memory_mb(in_mem))
        logging.info("   - Excitatory: %s", pretty_printing_memory_mb(ex_mem))
        logging.info(" - TOTAL : %s", pretty_printing_memory_mb(self.synapse_memory_total))

    @run_only_rank0
    def display_total(self):
        logging.info("+{:=^57}+".format(" Total Memory Estimates "))
        logging.info("| {:^40s} | {:^12s} |".format("Item", "Memory (MiB)"))
        logging.info("+{:-^57}+".format(""))
        full_overhead = self.base_memory * MPI.size
        logging.info("| {:<40s} | {:12.1f} |".format(f"Overhead (ranks={MPI.size})", full_overhead))
        logging.info("| {:<40s} | {:12.1f} |".format("Cells", self.cell_memory_total))
        logging.info("| {:<40s} | {:12.1f} |".format("Synapses", self.synapse_memory_total))
        logging.info("+{:-^57}+".format(""))
        grand_total = full_overhead + self.cell_memory_total + self.synapse_memory_total
        grand_total = pretty_printing_memory_mb(grand_total)
        logging.info("| {:<40s} | {:>12s} |".format("GRAND TOTAL", grand_total))
        logging.info("+{:-^57}+".format(""))

    def total_memory_available():
        """
        Returns the total memory available in the system in MB
        """
        try:
            virtual_memory = psutil.virtual_memory()
            return virtual_memory.total / (1024 * 1024)  # Total available memory in MB
        except Exception as e:
            logging.error(f"Error: {e}")
            return None

    @run_only_rank0
    def suggest_nodes(self, margin):
        """
        A function to calculate the suggested number of nodes to run the simulation
        The function takes into account the fact that the memory overhead is
        variable with the amount of ranks the simulation it's ran with.
        One can also specify a custom margin to add to the memory usage.
        """

        try:
            ranks_per_node = os.cpu_count()
        except AttributeError:
            ranks_per_node = multiprocessing.cpu_count()

        full_overhead = self.base_memory * ranks_per_node

        # initialize variable for iteration
        est_nodes = 0
        prev_est_nodes = None
        max_iter = 5
        iter_count = 0

        while (prev_est_nodes is None or est_nodes != prev_est_nodes) and iter_count < max_iter:
            prev_est_nodes = est_nodes
            mem_usage_per_node = full_overhead + self.cell_memory_total + self.synapse_memory_total
            mem_usage_with_margin = mem_usage_per_node * (1 + margin)
            est_nodes = math.ceil(mem_usage_with_margin / DryRunStats.total_memory_available())
            full_overhead = self.base_memory * ranks_per_node * est_nodes
            iter_count += 1

        return est_nodes

    @run_only_rank0
    def display_node_suggestions(self):
        """
        Display suggestions for how many nodes are approximately
        necessary to run the simulation based on the memory available
        on the current node.
        """
        node_total_memory = DryRunStats.total_memory_available()
        if node_total_memory is None:
            logging.warning("Unable to get the total memory available on the current node.")
            return
        suggested_nodes = self.suggest_nodes(0.3)
        logging.info(f"Based on the memory available on the current node, "
                     f"it is suggested to use at least {suggested_nodes} node(s).")
        logging.info("This is just a suggestion and the actual number of nodes "
                     "needed to run the simulation may be different.")
        logging.info(f"The calculation was based on a total memory available of "
                     f"{pretty_printing_memory_mb(node_total_memory)} on the current node.")
        logging.info("Please remember that it is suggested to use the same class of nodes "
                     "for both the dryrun and the actual simulation.")



"""
Stimuli sources. inc current and conductance sources which can be attached to cells
"""

from __future__ import absolute_import
from . import Neuron
from .random import RNG, gamma
import logging


class SignalSource:

    def __init__(self, base_amp=0.0, *, delay=0, rng=None):
        """
        Creates a new signal source, which can create composed signals
        Args:
            base_amp: The base (resting) amplitude of the signal (Default: 0)
            rng: The Random Number Generator. Used in the Noise functions
        """
        h = Neuron.h
        self.stim_vec = h.Vector()
        self.time_vec = h.Vector()
        self._cur_t = 0
        self._base_amp = base_amp
        self._rng = rng
        if delay > .0:
            self._add_point(base_amp)
            self._cur_t = delay

    def reset(self):
        self.stim_vec.resize(0)
        self.time_vec.resize(0)

    def _add_point(self, amp):
        """Appends a single point to the time-signal source.
        Note: It doesnt advance time, not supposed to be called directly
        """
        self.time_vec.append(self._cur_t)
        self.stim_vec.append(amp)

    def delay(self, duration):
        """Increments the ref time so that the next created signal is delayed
        """
        # NOTE: We rely on the fact that Neuron allows "instantaneous" changes
        # and made all signal shapes return to base_amp. Therefore delay() doesn't
        # need to introduce any point to avoid interpolation.
        self._cur_t += duration
        return self

    def add_segment(self, amp, duration, amp2=None):
        """Sets a linear signal for a certain duration.

        If amp2 is None (default) then we have constant signal
        """
        self._add_point(amp)
        self.delay(duration)
        self._add_point(amp if amp2 is None else amp2)
        return self

    def add_pulse(self, max_amp, duration, **kw):
        """Adds a pulse.

        A pulse is characterized by raising from a base amplitude, for a certain duration.
        """
        base_amp = kw.get("base_amp", self._base_amp)
        self._add_point(base_amp)
        self.add_segment(max_amp, duration)
        self._add_point(base_amp)
        return self

    def add_ramp(self, amp1, amp2, duration, **kw):
        """Adds a ramp.

        A ramp is characterized by a pulse whose peak changes uniformly during its length.
        Neuron automatically interpolates all values between [t0, t1] as a ramp
        """
        base_amp = kw.get("base_amp", self._base_amp)
        self._add_point(base_amp)
        self.add_segment(amp1, duration, amp2)
        self._add_point(base_amp)
        return self

    def add_train(self, amp, frequency, pulse_duration, total_duration, **kw):
        """Stimulus with repeated pulse injections at a specified frequency.

        Args:
            amp: the amplitude of a each pulse
            frequency: determines the number of pulses per second (hz)
            pulse_duration: the duration of a single pulse (peak time) (ms)
            total_duration: duration of the whole train (ms)
            base_amp: The base amplitude
        """
        base_amp = kw.get("base_amp", self._base_amp)
        tau = 1000 / frequency
        delay = tau - pulse_duration
        number_pulses = int(total_duration / tau)
        for _ in range(number_pulses):
            self.add_pulse(amp, pulse_duration, base_amp=base_amp)
            self.delay(delay)

        # Add final pulse, possibly partial
        remaining_time = total_duration - number_pulses * tau
        if pulse_duration <= remaining_time:
            self.add_pulse(amp, pulse_duration, base_amp=base_amp)
            self.delay(min(delay, remaining_time - pulse_duration))
        else:
            self.add_pulse(amp, remaining_time, base_amp=base_amp)
        # Last point
        self._add_point(base_amp)
        return self

    def add_sin(self, amp, total_duration, freq, step=0.025, **kw):
        """ Builds a sinusoidal signal.
        Args:
            amp: The max amplitude of the wave
            total_duration: Total duration, in ms
            freq: The wave frequency, in Hz
            step: The step, in ms (default: 0.025)
        """
        base_amp = kw.get("base_amp", self._base_amp)

        tvec = Neuron.h.Vector()
        tvec.indgen(self._cur_t, self._cur_t + total_duration, step)
        self.time_vec.append(tvec)
        self.delay(total_duration)

        stim = Neuron.h.Vector(len(tvec))
        stim.sin(freq, .0, step)
        stim.mul(amp)
        self.stim_vec.append(stim)
        self._add_point(base_amp)  # Last point
        return self

    def add_sinspec(self, start, dur):
        raise NotImplementedError()

    def add_pulses(self, pulse_duration, amp, *more_amps, **kw):
        """Appends a set of pulsed signals without returning to zero
           Each pulse is applied 'dur' time.

        Args:
          pulse_duration: The duration of each pulse
          amp: The amplitude of the first pulse
          *more_amps: 2nd, 3rd, ... pulse amplitudes
          **kw: Additional params:
            - base_amp [default: 0]
        """
        # First and last are base_amp
        base_amp = kw.get("base_amp", self._base_amp)
        self._add_point(base_amp)
        self.add_segment(amp, pulse_duration)
        for amp in more_amps:
            self.add_segment(amp, pulse_duration)
        self._add_point(base_amp)
        return self

    def add_noise(self, mean, variance, duration, dt=0.5):
        """Adds a noise component to the signal.
        """
        rng = self._rng or RNG()  # Creates a default RNG
        if not self._rng:
            logging.warning("Using a default RNG for noise generation")
        rng.normal(mean, variance)
        tvec = Neuron.h.Vector()
        tvec.indgen(self._cur_t, self._cur_t + duration, dt)
        svec = Neuron.h.Vector(len(tvec))
        svec.setrand(rng)

        # Delimit noise signals with base_amp
        # Otherwise Neuron does interpolation with surrounding points
        self._add_point(self._base_amp)
        self.time_vec.append(tvec)
        self.stim_vec.append(svec)
        self._cur_t += duration
        self._add_point(self._base_amp)
        return self

    def add_shot_noise(self, tau_D, tau_R, rate, amp_mean, amp_var, duration, dt=0.25):
        """
        Adds a Poisson shot noise signal with gamma-distributed amplitudes and
        bi-exponential impulse response.

        tau_D: bi-exponential decay time [ms]
        tau_R: bi-exponential rise time [ms]
        rate: Poisson event rate [Hz]
        amp_mean: mean of gamma-distributed amplitudes [nA]
        amp_var: variance of gamma-distributed amplitudes [nA^2]
        duration: duration of signal [ms]
        dt: timestep [ms]
        """
        from math import sqrt, exp, log

        rng = self._rng or RNG()  # Creates a default RNG
        if not self._rng:
            logging.warning("Using a default RNG for shot noise generation")

        tvec = Neuron.h.Vector()
        tvec.indgen(self._cur_t, self._cur_t + duration, dt)  # time vector
        ntstep = len(tvec)  # total number of timesteps

        rate_ms = rate / 1000  # rate in 1 / ms [mHz]
        napprox = 1 + int(duration * rate_ms)       # approximate number of events, at least one
        napprox = int(napprox + 3 * sqrt(napprox))  # better bound, as in elephant

        exp_scale = 1 / rate  # scale parameter of exponential distribution of time intervals
        rng.negexp(exp_scale)
        iei = Neuron.h.Vector(napprox)
        iei.setrand(rng)  # generate inter-event intervals

        ev = Neuron.h.Vector()
        ev.integral(iei, 1).mul(1000)  # generate events in ms
        # add events if last event falls short of duration
        while ev[-1] < duration:
            iei_new = Neuron.h.Vector(100)  # generate 100 new inter-event intervals
            iei_new.setrand(rng)            # here rng is still negexp
            ev_new = Neuron.h.Vector()
            ev_new.integral(iei_new, 1).mul(1000).add(ev[-1])  # generate new shifted events in ms
            ev.append(ev_new)  # append new events
        ev.where("<", duration)  # remove events exceeding duration
        ev.div(dt)  # divide events by timestep

        nev = Neuron.h.Vector([round(x) for x in ev])  # round to integer timestep index
        nev.where("<", ntstep)  # remove events exceeding number of timesteps

        sign = 1
        # if amplitude mean is negative, invert sign of current
        if amp_mean < 0:
            amp_mean = -amp_mean
            sign = -1

        gamma_scale = amp_var / amp_mean      # scale parameter of gamma distribution
        gamma_shape = amp_mean / gamma_scale  # shape parameter of gamma distribution
        # sample gamma-distributed amplitudes
        amp = gamma(rng, gamma_shape, gamma_scale, len(nev))

        E = Neuron.h.Vector(ntstep, 0)  # full signal
        for n, A in zip(nev, amp):
            E.x[int(n)] += sign * A  # add impulses, may overlap due to rounding to timestep

        # perform equivalent of convolution with bi-exponential impulse response
        # through a composite autoregressive process with impulse train as innovations

        # unitless quantities (time measured in timesteps)
        a = exp(-dt / tau_D)
        b = exp(-dt / tau_R)
        D = -log(a)
        R = -log(b)
        t_peak = log(R / D) / (R - D)
        A = (a / b - 1) / (a ** t_peak - b ** t_peak)

        P = Neuron.h.Vector(ntstep, 0)
        B = Neuron.h.Vector(ntstep, 0)

        # composite autoregressive process with exact solution
        # P[n] = b * (a ^ n - b ^ n) / (a - b)
        # for unit response B[0] = P[0] = 0, E[0] = 1
        for n in range(1, ntstep):
            P.x[n] = a * P[n - 1] + b * B[n - 1]
            B.x[n] = b * B[n - 1] + E[n - 1]

        P.mul(A)  # normalize to peak amplitude

        self._add_point(self._base_amp)
        self.time_vec.append(tvec)
        self.stim_vec.append(P)
        self._cur_t += duration
        self._add_point(self._base_amp)

        return self

    def add_ornstein_uhlenbeck(self, tau, sigma, mean, duration, dt=0.25):
        """
        Adds an Ornstein-Uhlenbeck process with given correlation time,
        standard deviation and mean value.

        tau: correlation time [ms], white noise if zero
        sigma: standard deviation [uS]
        mean: mean value [uS]
        duration: duration of signal [ms]
        dt: timestep [ms]
        """
        from math import sqrt, exp

        rng = self._rng or RNG()  # Creates a default RNG
        if not self._rng:
            logging.warning("Using a default RNG for Ornstein-Uhlenbeck process")

        tvec = Neuron.h.Vector()
        tvec.indgen(self._cur_t, self._cur_t + duration, dt)  # time vector
        ntstep = len(tvec)  # total number of timesteps

        svec = Neuron.h.Vector(ntstep, 0)  # stim vector

        noise = Neuron.h.Vector(ntstep)  # Gaussian noise
        rng.normal(0.0, 1.0)
        noise.setrand(rng)  # generate Gaussian noise

        if tau < 1e-9:
            svec = noise.mul(sigma)  # white noise
        else:
            mu = exp(-dt / tau)  # auxiliar factor [unitless]
            A = sigma * sqrt(1 - mu * mu)  # amplitude [uS]
            noise.mul(A)  # scale noise by amplitude [uS]

            # Exact update formula (independent of dt) from Gillespie 1996
            for n in range(1, ntstep):
                svec.x[n] = svec[n - 1] * mu + noise[n]  # signal [uS]

        svec.add(mean)  # shift signal by mean value [uS]

        self._add_point(self._base_amp)
        self.time_vec.append(tvec)
        self.stim_vec.append(svec)
        self._cur_t += duration
        self._add_point(self._base_amp)

        return self

    # PLOTTING
    def plot(self, ylims=None):
        from matplotlib import pyplot
        fig = pyplot.figure()
        ax = fig.add_subplot(1, 1, 1)  # (nrows, ncols, axnum)
        ax.plot(self.time_vec, self.stim_vec, label="Signal amplitude")
        ax.legend()
        if ylims:
            ax.set_ylim(*ylims)
        fig.show()

    # ==== Helpers =====
    # Helper methods forward generic kwargs to base class, like rng and delay

    @classmethod
    def pulse(cls, max_amp, duration, base_amp=.0, **kw):
        return cls(base_amp, **kw).add_pulse(max_amp, duration)

    @classmethod
    def ramp(cls, amp1, amp2, duration, base_amp=.0, **kw):
        return cls(base_amp, **kw).add_ramp(amp1, amp2, duration)

    @classmethod
    def train(cls, amp, frequency, pulse_duration, total_duration, base_amp=.0, **kw):
        return cls(base_amp, **kw).add_train(amp, frequency, pulse_duration, total_duration)

    @classmethod
    def sin(cls, amp, total_duration, freq, step=0.025, base_amp=.0, **kw):
        return cls(base_amp, **kw).add_sin(amp, total_duration, freq, step)

    @classmethod
    def noise(cls, mean, variance, duration, dt=0.5, base_amp=.0, **kw):
        return cls(base_amp, **kw).add_noise(mean, variance, duration, dt)

    @classmethod
    def shot_noise(cls, tau_D, tau_R, rate, amp_mean, var, duration, dt=0.25, base_amp=.0, **kw):
        return cls(base_amp, **kw).add_shot_noise(tau_D, tau_R, rate, amp_mean, var, duration, dt)

    @classmethod
    def ornstein_uhlenbeck(cls, tau, sigma, mean, duration, dt=0.25, base_amp=.0, **kw):
        return cls(base_amp, **kw).add_ornstein_uhlenbeck(tau, sigma, mean,duration, dt)

    # Operations
    def __add__(self, other):
        """# Adds signals. Two added signals sum amplitudes"""
        raise NotImplementedError("Adding signals is not available yet")


class CurrentSource(SignalSource):
    _all_sources = []

    def __init__(self, base_amp=0.0, *, delay=0, rng=None):
        """
        Creates a new current source that injects a signal under IClamp
        """
        super().__init__(base_amp, delay=delay, rng=rng)
        self._clamps = set()
        self._all_sources.append(self)

    class _Clamp:
        def __init__(self, cell_section, position=0.5, clamp_container=None,
                     stim_vec_mode=True, time_vec=None, stim_vec=None,
                     **clamp_params):
            self.clamp = Neuron.h.IClamp(position, sec=cell_section)
            if stim_vec_mode:
                assert time_vec is not None and stim_vec is not None
                self.clamp.dur = time_vec[-1]
                stim_vec.play(self.clamp._ref_amp, time_vec, 1)
            else:
                for param, val in clamp_params.items():
                    setattr(self.clamp, param, val)
            # Clamps must be kept otherwise they are garbage-collected
            self._all_clamps = clamp_container
            clamp_container.add(self)

        def detach(self):
            """Detaches a clamp from a cell, destroying it"""
            self._all_clamps.discard(self)
            del self.clamp  # Force del on the clamp (there might be references to self)

    def attach_to(self, section, position=0.5):
        return CurrentSource._Clamp(section, position, self._clamps, True,
                                    self.time_vec, self.stim_vec)

    # Constant has a special attach_to and doesnt share any composing method
    class Constant:
        """Class implementing a minimal IClamp for a Constant current."""
        _clamps = set()

        def __init__(self, amp, duration, delay=0):
            self._amp = amp
            self._dur = duration
            self._delay = delay

        def attach_to(self, section, position=0.5):
            return CurrentSource._Clamp(section, position, self._clamps, False,
                                        amp=self._amp, delay=self._delay, dur=self._dur)


class ConductanceSource(SignalSource):
    _all_sources = []

    def __init__(self, reversal=0.0, *, delay=.0, rng=None):
        """
        Creates a new conductance source that injects a conductance by driving
        the rs of an SEClamp at a given reversal potential.

        reversal: reversal potential of conductance (mV)
        """
        super().__init__(0.0, delay=delay, rng=rng)  # set SignalSource's base_amp to zero
        self._reversal = reversal   # set reversal from base_amp parameter in classmethods
        self._clamps = set()
        self._all_sources.append(self)

    class _DynamicClamp:
        def __init__(self, cell_section, position=0.5, clamp_container=None,
                     stim_vec_mode=True, time_vec=None, stim_vec=None,
                     reversal=0.0, **clamp_params):
            self.clamp = Neuron.h.SEClamp(position, sec=cell_section)
            if stim_vec_mode:
                assert time_vec is not None and stim_vec is not None
                self.clamp.dur1 = time_vec[-1]
                self.clamp.amp1 = reversal
                # support delay with initial zero
                self.time_vec = Neuron.h.Vector(1, 0).append(time_vec)
                self.stim_vec = Neuron.h.Vector(1, 0).append(stim_vec)
                # replace self.stim_vec with inverted and clamped signal
                # rs is in MOhm, so conductance is in uS (micro Siemens)
                self.stim_vec = Neuron.h.Vector(
                    [1 / x if x > 1E-9 and x < 1E9 else 1E9 for x in self.stim_vec])
                self.stim_vec.play(self.clamp._ref_rs, self.time_vec, 1)
            else:
                for param, val in clamp_params.items():
                    setattr(self.clamp, param, val)
            # Clamps must be kept otherwise they are garbage-collected
            self._all_clamps = clamp_container
            clamp_container.add(self)

        def detach(self):
            """Detaches a clamp from a cell, destroying it"""
            self._all_clamps.discard(self)
            del self.clamp  # Force del on the clamp (there might be references to self)

    def attach_to(self, section, position=0.5):
        return ConductanceSource._DynamicClamp(section, position, self._clamps, True,
                                               self.time_vec, self.stim_vec, self._reversal)


# EStim class is a derivative of TStim for stimuli with an extracelular electrode. The main
# difference is that it collects all elementary stimuli pulses and converts them using a
# VirtualElectrode object before it injects anything
#
# The stimulus is defined on the hoc level by using the addpoint function for every (step) change
# in extracellular electrode voltage. At this stage only step changes can be used. Gradual,
# i.e. sinusoidal changes will be implemented in the future
# After every step has been defined, you have to call initElec() to perform the frequency dependent
# transformation. This transformation turns e_electrode into e_extracellular at distance d=1 micron
# from the electrode. After the transformation is complete, NO MORE STEPS CAN BE ADDED!
# You can then use inject() to first scale e_extracellular down by a distance dependent factor
# and then vector.play() it into the currently accessed compartment
#
# TODO: 1. more stimulus primitives than step. 2. a dt of 0.1 ms is hardcoded. make this flexible!
